#include "collection.h"
#include <ne_dates.h>
#include <ne_basic.h>

namespace webdav
{

struct FetchContext
{
	FetchContext()
		: list(NULL)
		, target(NULL)
		, include_target(false)
	{}

	virtual ~FetchContext()
	{}

    Resource** list;
    const char* target;
    bool include_target;
};

static const ne_propname props[] = {
    { "DAV:", "getcontentlength" },
    { "DAV:", "getlastmodified" },
    { "http://apache.org/dav/props/", "executable" },
    { "DAV:", "resourcetype" },
    { "DAV:", "checked-in" },
    { "DAV:", "checked-out" },
    { NULL }
};

enum
{
	ELM_resourcetype = NE_PROPS_STATE_TOP + 1,
	ELM_collection,
};

static const struct ne_xml_idmap idmap[] = 
{
    { "DAV:", "resourcetype", ELM_resourcetype },
    { "DAV:", "collection", ELM_collection }
};

Collection::Collection(Session& session)
	: sess_(session)
{
}

Collection::~Collection()
{
}

bool Collection::GetList(const char* path,
						 const char* uri,
						 Resource** buf,
						 size_t* buf_size)
{
	*buf = NULL;
	*buf_size = 0;

	Resource *reslist = NULL;

	int res = FetchResourceList(uri, &reslist);
	if (res == NE_OK) {
		if (reslist == NULL) {
			// TRACE("collection is empty.\n");
		}else{
			Resource *current = NULL;
			Resource *next = NULL;
			for(current = reslist; current != NULL; current = next){
				next = current->Next();
				if (strlen(current->GetPath()) > strlen(path)) {
					(*buf_size)++;
				}
			}

			if(*buf_size == 0){
				return true;
			}

			*buf = new Resource[*buf_size];

			int i = 0;
			for(current = reslist; current != NULL; current = next){
				next = current->Next();
				if (strlen(current->GetPath()) > strlen(path)) {
					*(*buf + i) = *current;
					i++;
				}

				if(current != NULL){
					delete current;
					current = NULL;
				}
			}
		}

		return true;
	}

	return false;
}

int Collection::FetchResourceList(const char* uri, Resource **reslist)
{
	ne_propfind_handler* pfh = ne_propfind_create(sess_.GetSession(), uri, NE_DEPTH_ONE);
	FetchContext ctx;

	*reslist = NULL;

	ctx.list = reslist;
	ctx.target = uri;
	ctx.include_target = false;

	ne_xml_push_handler(ne_propfind_get_parser(pfh),
						StartElement,
						NULL,
						NULL,
						pfh);

	ne_propfind_set_private(pfh, CreatePrivate, NULL, NULL);

	int ret = ne_propfind_named(pfh, props, Results, &ctx);

	ne_propfind_destroy(pfh);

	return ret;
}

void Collection::Results(void *userdata, const ne_uri *uri,
						    const ne_prop_result_set *set)
{
	FetchContext* ctx = (FetchContext*)userdata;
	Resource* newres = (Resource*)ne_propset_private(set);

	if(uri->path == NULL){
		if(newres != NULL){
			delete newres;
		}

		return;
	}

	if(ne_path_compare(ctx->target, uri->path) == 0 && !ctx->include_target) {
		if(newres != NULL){
			delete newres;
		}

		return;
	}

	newres->SetPath(uri->path);

	const char* clength = ne_propset_value(set, &props[0]);    
	const char* modtime = ne_propset_value(set, &props[1]);
	const char* isexec = ne_propset_value(set, &props[2]);
	const char* checkin = ne_propset_value(set, &props[4]);
	const char* checkout = ne_propset_value(set, &props[5]);
	
	const ne_status *status = NULL;

	if(clength == NULL){
		status = ne_propset_status(set, &props[0]);
	}

	if(modtime == NULL){
		status = ne_propset_status(set, &props[1]);
	}

	if(newres->GetType() == Resource::TYPE_NORMAL && status != NULL){
		newres->SetErrorStatus(status->code);

		// Special hack for Apache 1.3/mod_dav
		if(strcmp(status->reason_phrase, "status text goes here") == 0){
			const char *desc;

			if(status->code == 401) {
				desc = "Authorization Required";
			}else if(status->klass == 3){
				desc = "Redirect";
			}else if(status->klass == 5){
				desc = "Server Error";
			}else{
				desc = "Unknown Error";
			}

			newres->SetErrorReason(desc);
		} else {
			newres->SetErrorReason(status->reason_phrase);
		}

		newres->SetType(Resource::TYPE_ERROR);
	}

	if(isexec && stricmp(isexec, "T") == 0) {
		newres->SetExecutable(true);
	} else {
		newres->SetExecutable(false);
	}

	if(modtime != NULL){
		newres->SetModifiedTime(ne_httpdate_parse(modtime));
	}

	if(clength != NULL){
		newres->SetSize(_atoi64(clength));
	}

	// o[WRg[
	if(checkin != NULL) {
		newres->SetVersionControl(1);
	} else if(checkout != NULL) {
		newres->SetVersionControl(2);
	} else {
		newres->SetVersionControl(0);
	}

	if(*ctx->list == NULL){
		*ctx->list = newres;
	}else{
		Resource* next = *ctx->list;
		*ctx->list = newres;
		(*ctx->list)->Insert(next);
	}
}

int Collection::StartElement(void *userdata, int parent, 
							   const char *nspace, const char *name, const char **atts)
{
	ne_propfind_handler* pfh = (ne_propfind_handler*)userdata;

	Resource* res = (Resource*)ne_propfind_current_private(pfh);

	int state = ne_xml_mapid(idmap, NE_XML_MAPLEN(idmap), nspace, name);

	if(res == NULL ||
		!((parent == NE_207_STATE_PROP && state == ELM_resourcetype) ||
		(parent == ELM_resourcetype && state == ELM_collection))){
		return NE_XML_DECLINE;
	}

	if(state == ELM_collection){
		res->SetType(Resource::TYPE_COLLECTION);
	}

	return state;
}

void* Collection::CreatePrivate(void *userdata, const ne_uri *uri)
{
	return new Resource;
}

} // namespace webdav
