#include "webdavimpl.h"
#include "collection.h"
#include <ne_basic.h>
#include "uri.h"
#include <sys/stat.h> /* for struct stat */
#include <ne_locks.h>
#include <ne_socket.h>
#include <io.h>
#include "http_put.h"
#include "http_get.h"
#include <tchar.h>

#include <Windows.h>
#include <Wincrypt.h>
#include <ne_string.h>

#pragma comment(lib, "crypt32.lib")

namespace webdav
{

WebDAVImpl::WebDAVImpl(WebDAV* iface)
	: interface_(iface)
	, last_status_(OK)
	, m_SslServerCert(NULL)
{
}

WebDAVImpl::~WebDAVImpl(void)
{
	if(ssl_certificate_.subject != NULL){
		::CertFreeCertificateContext(ssl_certificate_.subject);
		ssl_certificate_.subject = NULL;
	}

	if(m_SslServerCert != NULL){
		::CertFreeCertificateContext(m_SslServerCert);
		m_SslServerCert = NULL;
	}
}

void WebDAVImpl::SetUserName(const char* username)
{
	sess_.SetUserName(username);
}

void WebDAVImpl::SetPassword(const char* password)
{
	sess_.SetPassword(password);
}

void WebDAVImpl::SetProxy(const char* host, UINT port, const char* username, const char* password)
{
	sess_.SetProxyHost(host);
	sess_.SetProxyPort(port);
	sess_.SetProxyUserName(username);
	sess_.SetProxyPassword(password);
}

void WebDAVImpl::SetUserAgent(const char* token)
{
	sess_.SetUserAgent(token);
}

bool WebDAVImpl::IsConnected() const
{
	return sess_.IsConnected();
}

bool WebDAVImpl::Connect(const char* uri)
{
	URI uri2(uri);

	if(uri2.GetPort() == 0){
		Connect(uri2.GetScheme().c_str(), uri2.GetHost().c_str());
	}else{
		Connect(uri2.GetScheme().c_str(), uri2.GetHost().c_str(), uri2.GetPort());
	}

	if(!sess_.IsConnected()){
		return false;
	}

	return true;
}

bool WebDAVImpl::Connect(const char* scheme, const char* host, unsigned int port)
{
	sess_.Connect(scheme, host, port, NotifyProgress, NotifyStatus, VerifyServerCertificate, interface_);

	if(!sess_.IsConnected()){
		return false;
	}

	return true;
}

bool WebDAVImpl::Connect(const char* scheme, const char* host)
{
	sess_.Connect(scheme, host,  NotifyProgress, NotifyStatus, VerifyServerCertificate, interface_);

	if(!sess_.IsConnected()){
		return false;
	}

	return true;
}

bool WebDAVImpl::GetList(const char* path, Resource** buf, size_t* buf_size)
{
	if(!sess_.IsConnected()){
		return false;
	}

	std::string uri;

	uri = sess_.GetScheme() + _T("://") + sess_.GetURI().GetHost() + path;

	int index = 0;

	Collection collection(sess_);

	if(!collection.GetList(path, uri.c_str(), buf, buf_size)){
		return false;
	}

	return true;
}

bool WebDAVImpl::Get(const char* rpath, const char* lpath)
{
	if(!sess_.IsConnected()){
		return false;
	}

	HttpGet http;
	http.SetRequestHandler(NotifyRequest, interface_);
	http.SetResponseHandler(NotifyResponse, interface_);

	return http.Get(sess_.GetSession(), rpath, lpath);
}

bool WebDAVImpl::Put(const char* rpath, const char* lpath)
{
	if(!sess_.IsConnected()){
		return false;
	}

	HttpPut http;
	http.SetRequestHandler(NotifyRequest, interface_);
	http.SetResponseHandler(NotifyResponse, interface_);

	return http.Put(sess_.GetSession(), rpath, lpath);
}

bool WebDAVImpl::Copy(const char* spath, const char* dpath, bool is_overwrite)
{
	if(!sess_.IsConnected()){
		return false;
	}

	int ret = ne_copy(sess_.GetSession(), is_overwrite ? 1 : 0, NE_DEPTH_INFINITE, spath, dpath);
	if(ret != NE_OK){
		return false;
	}

	return true;
}

bool WebDAVImpl::Move(const char* spath, const char* dpath, bool is_overwrite)
{
	if(!sess_.IsConnected()){
		return false;
	}

	int ret = ne_move(sess_.GetSession(), is_overwrite ? 1 : 0, spath, dpath);
	if(ret != NE_OK){
		return false;
	}

	return true;
}

bool WebDAVImpl::MakeCollection(const char* path)
{
	if(!sess_.IsConnected()){
		return false;
	}

	int ret = ne_mkcol(sess_.GetSession(), path);
	if(ret != NE_OK){
		return false;
	}

	return true;
}

bool WebDAVImpl::Delete(const char* path)
{
	if(!sess_.IsConnected()){
		return false;
	}

	int ret = ne_delete(sess_.GetSession(), path);
	if(ret != NE_OK){
		return false;
	}

	return true;
}

void WebDAVImpl::NotifyStatus(void *userdata, ne_session_status status,
                              const ne_session_status_info *info)
{
	webdav::NotifyStatus st = NOTIFY_LOOKUP;
	webdav::NotifyInfo inf;

	(reinterpret_cast<WebDAV*>(userdata))->NotifyStatus(st, &inf);
}

void WebDAVImpl::NotifyProgress(void* userdata, off64_t progress, off64_t total)
{
	(reinterpret_cast<WebDAV*>(userdata))->NotifyProgress(progress, total);
}

int WebDAVImpl::NotifyResponse(void *userdata, const char *buffer, size_t buflen)
{
	return (reinterpret_cast<WebDAV*>(userdata))->NotifyResponse(buffer, buflen);
}

ssize_t WebDAVImpl::NotifyRequest(void *userdata,  char *buffer, size_t buflen)
{
	return (reinterpret_cast<WebDAV*>(userdata))->NotifyRequest(buffer, buflen);
}

int WebDAVImpl::VerifyServerCertificate(void *userdata, int failures, const ne_ssl_certificate *cert)
{
	int f = 0x00;
	SslCertificate ssl_cert;

	char* base64_cert = ne_ssl_cert_export(cert);
	unsigned char* unbase64_cert = NULL;
	size_t len = ne_unbase64(base64_cert, &unbase64_cert);
	free(base64_cert);

	PCCERT_CONTEXT cert_context = ::CertCreateCertificateContext(X509_ASN_ENCODING, unbase64_cert, len);
	free(unbase64_cert);
	
	WebDAV* webdav = (reinterpret_cast<WebDAV*>(userdata));

	if(webdav->pimpl_->ssl_certificate_.subject != NULL){
		::CertFreeCertificateContext(webdav->pimpl_->ssl_certificate_.subject);
		webdav->pimpl_->ssl_certificate_.subject = NULL;
	}

	webdav->pimpl_->ssl_certificate_.subject = cert_context;

	if(failures & NE_SSL_NOTYETVALID){
		f |= NotYetValid;
	}
	
	if(failures & NE_SSL_EXPIRED){
		f |= Expired;
	}
	
	if(failures & NE_SSL_IDMISMATCH){
		f |= IdMismatch;
	}
	
	if(failures & NE_SSL_UNTRUSTED){
		f |= Untrusted;
	}
	
	if(failures & NE_SSL_BADCHAIN){
		f |= BadChain;
	}
	
	if(failures & NE_SSL_REVOKED){
		f |= Revoked;
	}

	return webdav->VerifyServerCertificate(f, &(webdav->pimpl_->ssl_certificate_));
}

static int all_iterator(void *userdata,
						const ne_propname *pname,
						const char *value,
						const ne_status *status)
{
    return 0;
}

static void pget_results(void *userdata, 
						 const ne_uri *uri,
						 const ne_prop_result_set *set)
{
	ne_propset_iterate(set, all_iterator, NULL);

	return;
}

bool WebDAVImpl::GetFileProperty(const char* path)
{
	if(!sess_.IsConnected()){
		return false;
	}

	ne_session *sess = sess_.GetSession();

	int ret = ne_simple_propfind(sess,
								 path,
								 NE_DEPTH_ZERO,
								 NULL,
								 pget_results,
								 NULL);

	if(ret == NE_OK){
		return true;
	}

	return false;
}

PCCERT_CONTEXT WebDAVImpl::GetSslServerCert()
{
	ne_ssl_certificate* cert = sess_.GetSslServerCert();
	if(cert == NULL){
		return NULL;
	}

	char* base64_cert = ne_ssl_cert_export(cert);
	unsigned char* unbase64_cert = NULL;
	size_t len = ne_unbase64(base64_cert, &unbase64_cert);
	free(base64_cert);

	if(m_SslServerCert != NULL){
		::CertFreeCertificateContext(m_SslServerCert);
		m_SslServerCert = NULL;
	}

	m_SslServerCert = ::CertCreateCertificateContext(X509_ASN_ENCODING, unbase64_cert, len);
	free(unbase64_cert);

	if(m_SslServerCert == NULL){
		return NULL;
	}

	return m_SslServerCert;
}

}
